#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  6 09:32:12 2020

@author: benjamin
"""

import copy
import numpy as np
import scipy.interpolate as scintp
from .plottools import plot_area
from .utils import lininterp1

class AreaFunction:
    """class for manipulating area fucntions"""
    area = None
    length = None
    parent = None
    total_length = None
    constriction_area = None
    constriction_location = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            
        if self.area is not None and self.length is not None:
            _, _, _ = self.characteristics()

    def copy(self):
        """returns a deep copy of the instance"""
        return copy.deepcopy(self)

    def interpolate(self, num_tubes):
        """interpolates area function into a vector y_out of N_out evenly spaced area function."""
        
        af = self.area
        lf = self.length
        if len(af.shape) == 1:
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)

        cumx_in = np.cumsum(lf, 0)
        x_out = np.zeros((num_tubes, af.shape[1]))
        y_out = np.zeros((num_tubes, af.shape[1]))
        for kframe in range(af.shape[1]):
            y_in = af[:, kframe]
            x_in = cumx_in[:, kframe]

            x_out1 = np.arange(1, num_tubes+1)*x_in[-1]/num_tubes
            f1 = scintp.interp1d(x_in, y_in,
                                 kind='nearest', fill_value='extrapolate')
            y_out[:, kframe] = f1(x_out1)
            x_out[:, kframe] = np.append(x_out1[0], np.diff(x_out1))

        self.area = y_out.squeeze()
        self.length = x_out.squeeze()
        return x_out, y_out

    def plot(self, sl=None, hf=None):
        """plots the area function"""
        a = np.array(self.area)
        l = np.array(self.length)
        if sl is not None:
            a = a[:, sl]
            l = l[:, sl]
            
        xval = np.insert(np.cumsum(l) * 1e2, 0, 0)
        yval = np.insert(a * 1e4, 0, a[0] * 1e4)

        return plot_area(xval, yval, hf)

    def changejawopening(self, newangle, orig):
        """modifies the area function to simulate a modification of the jaw opening"""

        af = self.area
        lf = self.length

        xf = np.cumsum(lf, 0)
        if len(af.shape) == 1:
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)
            xf = xf.reshape(-1, 1)

        d_wall = np.sqrt((af[-1, :] - af[orig, :])**2 + (xf[-1, :] - xf[orig, :])**2)
        current_angle = np.sign(af[-1, :] - af[orig, :]) * np.arccos((xf[-1, :] - xf[orig, :]) / d_wall)
        anglechange = newangle - current_angle

        newxf2change = (xf[orig:, :] - xf[orig, :]) * np.cos(anglechange) - \
            (af[orig:, :] - af[orig, :]) * np.sin(anglechange)
        newaf2change = (xf[orig:, :] - xf[orig, :]) * np.sin(anglechange) + \
            (af[orig:, :] - af[orig, :]) * np.cos(anglechange)

        af[orig:, :] = newaf2change + af[orig, :]
        xf[orig:, :] = newxf2change + xf[orig, :]
        newlf = np.matlib.repmat(xf[-1, :] / af.shape[0], af.shape[0], af.shape[1])

        f1 = scintp.interp1d(xf.squeeze(), af.squeeze(),
                             kind='linear', fill_value='extrapolate')
        newaf = f1(np.cumsum(newlf, 0))

        if any(newaf <= 1e-11):
            print('Warning : contact detected!')
            newaf[newaf <= 1e-11] = 1e-11

        self.area = newaf.squeeze()
        self.length = newlf.squeeze()
        _, _, _ = self.characteristics()

        return newaf.squeeze(), newlf.squeeze()
    
    def changeprotrusion(self, new_length, orig_teeth=None):
        """modifies the area function to simulate a change in lip protrusion"""
        
        af = self.area
        lf = self.length

        if len(af.shape) == 1:
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)
            
        num_lips = af.shape[0] - orig_teeth - 1
        lf[orig_teeth+1:] = lf[orig_teeth+1:] + (new_length - np.sum(lf, 0)) / num_lips
        
        self.length = lf.squeeze()
        self.area = af.squeeze()
        _, _, _ = self.characteristics()
        
        return af.squeeze(), lf.squeeze()
    
    def areamorph(self, area_target, pct_path=50):
        """modifies the area function so that it goes a certain % towards a target"""
        
        
        afMat = np.hstack((self.area.reshape(-1, 1), area_target.area.reshape(-1, 1)))
        lfMat = np.hstack((self.length.reshape(-1, 1), area_target.length.reshape(-1, 1)))

        xi = [0, 1]
        xo = np.arange(101)/ 100

        afOut = lininterp1( xi, afMat, xo, 1 )
        lfOut = lininterp1( xi, lfMat, xo, 1)
        self.area = afOut[:,int(pct_path)]
        self.length = lfOut[:,int(pct_path)]
        
    
    def characteristics(self):
        """returns the characteristics of the area function"""
        
        self.total_length = np.sum(self.length, 0)
        self.constriction_area = np.min(self.area, 0)
        self.constriction_location = np.argmin(self.area, 0)
        
        return self.total_length, self.constriction_area, self.constriction_location


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 18 12:17:48 2021

@author: benjamin
"""

import os
from sklearn.mixture import GaussianMixture
from sklearn import linear_model
import h5py
import numpy as np
import multiprocessing
from joblib import Parallel, delayed
import struct
from feattools import get_delta

def db(x):
    return 20 * np.log10(np.abs(x))

def cc2ceps(x, nwin):
    L = int(nwin / 2 + 1)
    return 10**(np.real(np.fft.fft(x, nwin, axis=0))/20)[:L,:]


def plot_mfcc(mfcc):
    import matplotlib.pyplot as plt
    S = mfcc.mfcc2melspectro()
    plt.figure()
    plt.imshow(np.log10(S.time_frequency_matrix), aspect='auto', origin='lower')

def read_list(list_file):
    fid = open(list_file, 'r')
    list_smpl = fid.readlines()
    fid.close()
    return list_smpl

def get_spk_list(list_smpl):
    spk_id = []
    for spk in list_smpl:
        spk_tmp = spk.split(' ')[0]
        if spk_tmp not in spk_id:
            spk_id.append(spk_tmp)
    return list(np.sort(spk_id))

def get_audio_list(list_smpl, spk):
    list_wom = [x.split(' ')[1] for x in list_smpl if x.split(' ')[0] == spk]
    list_m = [x.split(' ')[2].replace('\n', '') for x in list_smpl if x.split(' ')[0] == spk]

    return list_wom, list_m

def get_spk_id(x):
    file_name = os.path.split(x)[-1]
    return os.path.splitext(file_name)[0]

def get_features(list_audio, path=None, sro=8000,
                 nwin=256, novlp=int(256*3/4), norm='std', axis=0):
    feat_mtx = None
    for audio in list_audio:
        if path is not None:
            audio_file = os.path.join(path, audio)
        else:
            audio_file = audio
        Sp = file2speech(audio_file, sro=sro)
        X, fxx, txx = Sp.computecepstrumenvelope(nwin=nwin, novlp=novlp)
        feat_mtx = concat_mtx(feat_mtx, X)
    return norm_feat(feat_mtx, axis=axis, method=norm)

def get_cc_features(list_audio, nord=32, path=None, sro=8000, nwin=256,
                    novlp=int(256*3/4), norm='zscore', axis=0):

    feat_mtx = None
    for audio in list_audio:
        if path is not None:
            audio_file = os.path.join(path, audio)
        else:
            audio_file = audio
        Sp = file2speech(audio_file, sro=sro)
        X, fxx, txx = Sp.computecepstrumenvelope(nwin=nwin, novlp=novlp)
        L = X.shape[0]
        nfft = 2 * (L - 1)
        X_cc = symmetricifft(db(X), nfft)[:nord,:]
        feat_mtx = concat_mtx(feat_mtx, X_cc)
    return norm_feat(feat_mtx, axis=axis, method=norm)

def get_mfcc_features(list_audio, path=None, sro=8000,
                 nwin=256, novlp=int(256*3/4), norm='zscore',
                 nb_filter=24, nb_coeff=12, axis=0, replace=False):
    feat_mtx = None
    c0 = None
    for audio in list_audio:
        if path is not None:
            audio_file = os.path.join(path, audio)
        else:
            audio_file = audio
        Sp = file2speech(audio_file, sro=sro)
        Sp.mfcc(nwin=nwin, novlp=novlp, nb_filter=nb_filter,
                nb_coeff=nb_coeff, preemph=[1, -0.95],
                win='hamming', replace=replace)
        X = Sp.mfcc.coeff
        if replace:
            c0 = concat_mtx(c0, X[0,:].reshape(1,-1))
        feat_mtx = concat_mtx(feat_mtx, X)
    return (norm_feat(feat_mtx, axis=axis, method=norm), c0)

def norm_feat(A, axis=0, method='std'):

    if method not in ['zscore', 'mean', 'std', 'max']:
        return A

    if method in ['zscore', 'mean']:
        mean_A = np.mean(A, axis=axis, keepdims=True)
    else:
        mean_A = 0
    A = A - mean_A
    if method in ['zscore', 'std']:
        return A / np.std(A, axis=axis, keepdims=True)
    elif method=='max':
        return A / np.max(A, axis=axis, keepdims=True)
    else:
        return A

def concat_mtx(x, y, axis=1):
    if x is None:
        x = y
    elif len(x.shape) == 1:
        x = np.vstack((x.reshape(-1, 1),
                       y.reshape(-1, 1)))
    elif axis == 1:
        x = np.hstack((x,y))
    elif axis == 0:
        x = np.vstack((x,y))
    return x.squeeze()

def gmm(features, nb_gauss):
    clf = GaussianMixture(n_components=nb_gauss, covariance_type='full')
    clf.fit(features)
    return clf

def regul_res(A_source, A_target):
    nb_feat, nb_ph = A_source.shape
    Tk = []
    for k in range(nb_ph):
        As_tmp = A_source[:,k].reshape(-1,1)
        At_tmp = A_target[:,k].reshape(-1,1)
        Tk.append(At_tmp @ np.linalg.pinv(As_tmp))

    return Tk

def trans_oper(A_source_norm, A_target):

    nb_ph, nb_feat = A_source_norm.shape
    Tk = []

    for k in range(nb_ph):
        As_tmp = A_source_norm[k,:].reshape(-1)
        At_tmp = A_target[k,:].reshape(-1)
        Tk.append(At_tmp / As_tmp)

    return Tk

def warp_res(A_targ, W_s, T_k, R_s):
    nb_feat, nb_ph = A_targ.shape
    nb_frame = R_s.shape[1]
    X_t = np.zeros_like(R_s)
    T_k.append(np.eye(nb_feat))
    W_plus = np.vstack((W_s, np.ones((1, nb_frame))))

    for k_fr in range(nb_frame):
        s = 0
        for k_ph in range(nb_ph + 1):
            s += W_plus[k_ph, k_fr] * T_k[k_ph]
        X_t[:, k_fr] = A_targ @ W_s[:, k_fr] + s @ R_s[:, k_fr]

    return X_t

def read_cep(file_name, details=False):
    
    with open(file_name, 'rb') as f:
        file_read = f.read()

    nb_utt = struct.unpack('i', file_read[:4])[0]
    nb_feat = struct.unpack('i', file_read[4:8])[0]
    
    fmat_one = 'f'
    fmat = fmat_one * nb_feat
    nb_byte = 4 * nb_feat
    if fmat_one == 'd':
        nb_byte *= 2
    
    imax = len(file_read)
    i0 = 8 + 4 * nb_utt
    i1 = i0 + nb_byte
    features = []
    while i1 <= imax:
        features.append(struct.unpack(fmat, file_read[i0:i1]))
        i0 += nb_byte
        i1 = i0 + nb_byte
        
    if details:
        nb_frame = struct.unpack('i'*nb_utt, file_read[8:8+4*nb_utt])
        return np.array(features).T, nb_frame
    else:
        return np.array(features).T

def write_cep(file_name, features, nb_frame, nb_utt=None, nb_feat=None):
    if nb_feat is None:
        nb_feat = features.shape[1]
    if nb_utt != len(nb_frame):
        nb_utt = len(nb_frame)
        
    f = open(file_name, 'w+b')
    l1 = struct.pack('i', nb_utt)
    l2 = struct.pack('i', nb_feat)
    f.write(l1)
    f.write(l2)
    
    for curr_frame in nb_frame:
        l3 = struct.pack('i', curr_frame )
        f.write(l3)
    for k1 in range(features.shape[0]):
        for k2 in range(features.shape[1]):
            l4 = struct.pack('f', features[k1, k2])
            f.write(l4)
    f.close()

def get_cep_features(list_cep, path=None):

    feat_mtx = None
    for cep in list_cep:
        if path is not None:
            cep_file = os.path.join(path, cep)
        else:
            cep_file = cep
        X = read_cep(cep_file, details=False)
        feat_mtx = concat_mtx(feat_mtx, X)
    return feat_mtx

def save_spk_model(gmm_model, a_mask, output_name, nd=2):

    hf = h5py.File(output_name, 'w')
    hf.create_dataset('means_nominal', data=gmm_model.means_)
    hf.create_dataset('weights', data=gmm_model.weights_)
    hf.create_dataset('cov', data=gmm_model.covariances_)
    hf.create_dataset('prec', data=gmm_model.precisions_)
    hf.create_dataset('prec_chol', data=gmm_model.precisions_cholesky_)
    hf.create_dataset('means_mask', data=a_mask)
    hf.create_dataset('nb_delta', data=nd)

    hf.close()

def read_spk_model(fileName):

    f = h5py.File(fileName, 'r')
    a_mask = f['means_mask'][()]
    nb_gauss, nb_feat = a_mask.shape
    gmm_model = GaussianMixture(n_components=nb_gauss, covariance_type='full')
    gmm_model.means_ = f['means_nominal'][()]
    gmm_model.weights_ = f['weights'][()]
    gmm_model.covariances_ = f['cov'][()]
    gmm_model.precisions_ = f['prec'][()]
    gmm_model.precisions_cholesky_ = f['prec_chol'][()]

    f.close()
    return gmm_model, a_mask

def read_spk_model_list(file_list):

    A_wom = None
    A_weights = None
    A_cov = None
    A_prec = None
    A_prec_chol = None
    A_m = None
    for fileName in file_list:   
        f = h5py.File(fileName, 'r')
        A_wom = concat_mtx(A_wom, f['means_nominal'][()], axis=0)        
        A_m = concat_mtx(A_m, f['means_mask'][()], axis=0)
        A_weights = concat_mtx(A_weights, f['weights'][()], axis=0)
        A_cov = concat_mtx(A_cov, f['cov'][()], axis=0)
        A_prec = concat_mtx(A_prec, f['prec'][()], axis=0)
        A_prec_chol = concat_mtx(A_prec_chol, f['prec_chol'][()], axis=0)
        nb_gauss, nb_feat = A_wom.shape

    gmm_model = GaussianMixture(n_components=nb_gauss, covariance_type='full')
    gmm_model.means_ = A_wom
    gmm_model.weights_ = A_weights
    gmm_model.covariances_ = A_cov
    gmm_model.precisions_ = A_prec
    gmm_model.precisions_cholesky_ = A_prec_chol

    f.close()
    return gmm_model, A_m

def transform_ceps(gmm_model, features, W_mask, norm='std', axis=0):

    A_source = gmm_model.means_    
    Tk_lin = trans_oper(A_source, W_mask)
    nb_feat, nb_frame = features.shape

    idx_test_atoms = gmm_model.predict(norm_feat(features.T, axis=axis,
                                                            method=norm))
    feat_out = np.zeros_like(features)
    for k in range(nb_frame):
        feat_out[:, k] = Tk_lin[idx_test_atoms[k]] * features[:, k]

    return feat_out

def transform_ceps_sparse(W_nominal, features, W_mask, norm='std', reg='res_only',
                           alpha=0.5, tol=1e-5, max_iter=100000, positive=False):

    clf = linear_model.Lasso(alpha=alpha, tol=tol, max_iter=max_iter,
                             positive=positive)
    
    clf.fit(W_nominal.T, features)
    W = clf.coef_.T
     
    X_nom = W_nominal.T @ W
    Rs = features - X_nom
    W_raw = W_mask.T @ W
    W_res = W_raw + Rs
    Tk = regul_res(W_nominal.T, W_mask.T)  
    W_full = warp_res(W_mask.T, W, Tk, Rs)
    if reg == 'full_all':
        return [W_raw, W_res, W_full]    
    elif reg == 'full_only':
        return W_full
    elif reg == 'res_only':
        return W_res
    elif reg == 'raw_only':
        return W_raw
    else:
        raise ValueError("Wrong method for regularization")

def display_speakers(spk_id, spk_done):
    if spk_done is None:
        l_spk = 'Processing ' + str(len(spk_id)) + ' speakers: '
        for spk in spk_id:
            l_spk += spk + ' '

        print(l_spk[:-1])
    else:
        spk_new_list = [x for x in spk_id if x not in spk_done]
        l_spk = 'Processing ' + str(len(spk_new_list)) + ' remaining speakers: '
        for spk in spk_new_list:
            l_spk += spk + ' '

        print(l_spk[:-1])

def coeff2mfcc(coeff, sro=8000, nb_filter=24, nb_coeff=12, nfft=1024, c0=None):

    filters, mels, mel_freqs = mel_filter_bank(sr=sro, fmin=0, fmax=sro/2,
                                               nb_filter=nb_filter, nfft=nfft)
    if c0 is None:
        c0 = np.array([x for x in coeff[0,:]])
    mfcc_object = MFCC('coeff', coeff, 'frequency_vector', mel_freqs,
                       'time_vector', np.arange(coeff.shape[1]) / sro,
                       'mel_vector', mels, 'nb_coeff', nb_coeff, 'c0', c0)

    return mfcc_object

def comp_list(a, b):
    if len(a) != len(b):
        return False
    for k1, k2 in zip(a, b):
        if k1 != k2:
            return False
        
    return True

def convert_cep_file(cep_file, file_list, method='sparse'):
    gmm_model, W_mask = read_spk_model_list(file_list)
    features = get_cep_features([cep_file], path=None)
    if method == 'sparse':
        return transform_ceps_sparse(gmm_model.means_, features,
                                         W_mask, norm=None)

    elif method == 'linear':
        return transform_ceps(gmm_model, features, W_mask, norm=None, axis=1)
    
def convert_cep_files(cep_list, file_list, method='sparse', path=None,
                      reg='full_all', nd=0):
    gmm_model, W_mask = read_spk_model_list(file_list)
    features = get_delta(get_cep_features(cep_list, path=path).T, nd).T
    if method == 'sparse':
        return transform_ceps_sparse(gmm_model.means_, features, W_mask, 
                                     norm=None, reg=reg)
    elif method == 'linear':
        return transform_ceps(gmm_model, features, W_mask, norm=None)

def convert_audio(audio_file, file_list, sro=8000, nwin=256, novlp=int(256*3/4),
                   norm=None, method='sparse', nb_filter=24, nb_coeff=12,
                  feat_type='ceps', chn=1, axis=0, replace=False):

    Sp = file2speech(audio_file, chn=chn, sro=sro)
    gmm_model, W_mask = read_spk_model_list(file_list)

    if feat_type == 'ceps':
        features = get_features([audio_file], path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, axis=axis)
        norm = 'std'
    if feat_type == 'log_ceps':
        features = get_features([audio_file], path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=None, axis=axis)
        features = np.log(features)
        norm = 'zscore'
    elif feat_type == 'cc':
        nb_coeff = W_mask.shape[1]
        features = get_cc_features([audio_file], nord=nb_coeff, path=None,
                           sro=sro, nwin=nwin, novlp=novlp, norm=norm, axis=axis)
        norm = 'zscore'
    elif feat_type == 'mfcc':
        features, c0 = get_mfcc_features([audio_file], path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, nb_filter=nb_filter,
                            nb_coeff=nb_coeff, axis=axis, replace=replace)
        norm = 'zscore'
    if method == 'sparse':
        feat_out = transform_ceps_sparse(gmm_model.means_, features,
                                         W_mask, norm=norm)

    elif method == 'linear':
        feat_out = transform_ceps(gmm_model, features, W_mask, norm=norm,
                                  axis=axis)
    else:
        raise ValueError('Method is not correct. Choose either sparse or linear')

    if feat_type in ['log_ceps', 'ceps', 'cc']:
        if feat_type == 'cc':
            feat_out = cc2ceps(feat_out, nwin)
        elif feat_type == 'log_ceps':
            feat_out = np.exp(feat_out)        
        return Sp.cepstraltransform(feat_out, nwin=nwin, novlp=novlp,
                                method='target')
    elif feat_type == 'mfcc':
        mfcc_out = coeff2mfcc(feat_out, sro=sro, nb_filter=nb_filter,
                              nb_coeff=nb_coeff, nfft=1024, c0=c0)
        return Sp.cepstraltransform(mfcc_out, nwin=nwin, novlp=novlp,
                                method='target', feat_type='mfcc')

def extract_speaker(spk, list_smpl, spk_done, spk_id, sro, nwin, novlp,
                    nb_gauss, verb, ext_name, feat_type, nb_filter, nb_coeff,
                    axis, replace):
    list_wom, list_m = get_audio_list(list_smpl, spk)

    if verb:
        display_speakers(spk_id, spk_done)
        print('Extracting features of ' + spk + '...')
    if feat_type == 'ceps':
        norm = 'std'
        feat_wom = get_features(list_wom, path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, axis=axis)
        feat_m = get_features(list_m, path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, axis=axis)
    elif feat_type == 'log_ceps':
        norm = None
        feat_wom = get_features(list_wom, path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, axis=axis)
        feat_m = get_features(list_m, path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, axis=axis)
        feat_wom = norm_feat(np.log(feat_wom), axis=axis, method='zscore')
        feat_m = norm_feat(np.log(feat_m), axis=axis, method='zscore')

    elif feat_type == 'mfcc':
        norm = 'zscore'
        feat_wom, c0 = get_mfcc_features(list_wom, path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, nb_filter=nb_filter,
                            nb_coeff=nb_coeff, axis=axis, replace=replace)
        feat_m, c0 = get_mfcc_features(list_m, path=None, sro=sro, nwin=nwin,
                            novlp=novlp, norm=norm, nb_filter=nb_filter,
                            nb_coeff=nb_coeff, axis=axis)
    elif feat_type == 'cc':
        norm = 'zscore'
        feat_wom = get_cc_features(list_wom, nord=nb_coeff, path=None,
                           sro=sro, nwin=nwin, novlp=novlp, norm=norm, axis=axis)
        feat_m = get_cc_features(list_m, nord=nb_coeff, path=None,
                           sro=sro, nwin=nwin, novlp=novlp, norm=norm, axis=axis)

    if verb:
        print('Extracting features of ' + spk + ' done')
        print('Estimating ' + str(nb_gauss) + ' gaussians for ' + spk + ' without mask...')

    clf_wom = gmm(feat_wom.T, nb_gauss)

    if verb:
        print('Estimating ' + str(nb_gauss) + ' gaussians for ' + spk + ' without mask done')

    idx_atoms = clf_wom.predict(feat_m.T)
    n_gauss_final = len(np.unique(idx_atoms))

    clf_wom.n_components = n_gauss_final
    if verb:
        print('Keeping ' + str(n_gauss_final) + ' gaussians out of ' + str(nb_gauss))
    clf_wom.means_ = clf_wom.means_[np.unique(idx_atoms),:]
    clf_wom.weights_ = clf_wom.weights_[np.unique(idx_atoms)]
    clf_wom.covariances_ = clf_wom.covariances_[np.unique(idx_atoms),:]
    clf_wom.precisions_ = clf_wom.precisions_[np.unique(idx_atoms),:]
    clf_wom.precisions_cholesky_ = clf_wom.precisions_cholesky_[np.unique(idx_atoms),:]

    A_wom = clf_wom.means_
    A_m = np.zeros_like(A_wom)

    for k_idx, idx in enumerate(np.unique(idx_atoms)):
        X = np.array([feat_m[:, x] for x in idx_atoms if x == idx])
        A_m[k_idx, :] = np.mean(X, 0)

    output_name = os.path.join(ext_name, spk + '.h5')
    save_spk_model(clf_wom, A_m, output_name)
    print('Dictionary file for ' + spk + ' has been successfully saved')
    if spk_done is None:
        spk_done = [spk]
    else:
        spk_done.append(spk)
        
        
def extract_speaker_2(spk, list_smpl_nomask, list_smpl_mask, spk_done, spk_id,
                    nb_gauss, verb, ext_name, nd):
    
    list_wom = [x for x in list_smpl_nomask if get_spk_id(x) == spk]
    list_m = [x for x in list_smpl_mask if get_spk_id(x) == spk]

    if verb:
        display_speakers(spk_id, spk_done)
        print('Extracting features of ' + spk + '...')
    
    feat_wom = get_delta(get_cep_features(list_wom, path=None).T, nd).T
    feat_m = get_delta(get_cep_features(list_m, path=None).T, nd).T
   
    if verb:
        print('Extracting features of ' + spk + ' done')
        print('Estimating ' + str(nb_gauss) + ' gaussians for ' + spk + ' without mask...')

    clf_wom = gmm(feat_wom.T, nb_gauss)

    if verb:
        print('Estimating ' + str(nb_gauss) + ' gaussians for ' + spk + ' without mask done')
    
    # clf_wom.means_ = (clf_wom.means_ , 1, 'zscore')
    idx_atoms = clf_wom.predict(feat_m.T)
    n_gauss_final = len(np.unique(idx_atoms))

    clf_wom.n_components = n_gauss_final
    if verb:
        print('Keeping ' + str(n_gauss_final) + ' gaussians out of ' + str(nb_gauss))
    clf_wom.means_ = clf_wom.means_[np.unique(idx_atoms),:]
    clf_wom.weights_ = clf_wom.weights_[np.unique(idx_atoms)]
    clf_wom.covariances_ = clf_wom.covariances_[np.unique(idx_atoms),:]
    clf_wom.precisions_ = clf_wom.precisions_[np.unique(idx_atoms),:]
    clf_wom.precisions_cholesky_ = clf_wom.precisions_cholesky_[np.unique(idx_atoms),:]

    A_wom = clf_wom.means_
    A_m = np.zeros_like(A_wom)

    for k_idx, idx in enumerate(np.unique(idx_atoms)):
        X = np.array([feat_m[:, x] for x in idx_atoms if x == idx])
        A_m[k_idx, :] = np.mean(X, 0)

    output_name = os.path.join(ext_name, spk + '.h5')
    
    save_spk_model(clf_wom, A_m, output_name, nd)
    print('Dictionary file for ' + spk + ' has been successfully saved')
    if spk_done is None:
        spk_done = [spk]
    else:
        spk_done.append(spk)

def extract_dict(list_file, nb_gauss, sro, nwin, novlp, ext_name, njobs, verb,
                  feat_type, nb_filter, nb_coeff, axis, replace):
    list_smpl = read_list(list_file)
    spk_id = get_spk_list(list_smpl)
    spk_done = None

    njobs = min(njobs, len(spk_id))
    print('Expected number of parallel jobs is ' + str(njobs))
    if njobs > 1:
        verb = False
        display_speakers(spk_id, spk_done)

    if not os.path.isdir(ext_name):
        os.mkdir(ext_name)

    pList = Parallel(n_jobs=njobs)(delayed(extract_speaker)(spk,
                          list_smpl, spk_done, spk_id, sro, nwin, novlp,
                    nb_gauss, verb, ext_name, feat_type,
                    nb_filter, nb_coeff, axis, replace) for spk in spk_id)
    print('All dictionaries have been successfully saved')
    
def extract_dict_2(nomask_set, mask_set, nb_gauss, ext_name, njobs, nd, verb, spk):
    
    list_cep_nomask = [os.path.join(nomask_set, x) for x in np.sort(os.listdir(nomask_set))
                       if '.cep' in x]
    list_cep_mask = [os.path.join(mask_set, x) for x in np.sort(os.listdir(mask_set))
                       if '.cep' in x]
    
    spk_id_nom = [os.path.splitext(x)[0] for x in np.sort(os.listdir(nomask_set)) if '.cep' in x]
    spk_id_mask = [os.path.splitext(x)[0] for x in np.sort(os.listdir(mask_set)) if '.cep' in x]
    
    if not comp_list(spk_id_nom, spk_id_mask):
        raise ValueError('Speaker lists are different. ' + 
                         'Please ensure speakers are the same for both mask and non-mask lists')
    spk_id = get_spk_list(spk_id_nom)
    if spk is not None:
        spk_id = [x for x in spk_id if x in spk]
    spk_done = None
    njobs = min(njobs, len(spk_id))
    print('Expected number of parallel jobs is ' + str(njobs))
    if njobs > 1:
        verb = False
        display_speakers(spk_id, spk_done)

    if not os.path.isdir(ext_name):
        os.mkdir(ext_name)

    pList = Parallel(n_jobs=njobs)(delayed(extract_speaker_2)(spk,
                          list_cep_nomask, list_cep_mask, spk_done, spk_id,
                    nb_gauss, verb, ext_name, nd) for spk in spk_id)
    print('All dictionaries have been successfully saved')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 27 21:07:26 2020

@author: benjamin
"""

import copy
import numpy as np
import statsmodels.api as sm
import scipy.interpolate as scintp
import warnings
from .plottools import pitch_plot, contour_plot
from .splinetools import (check_list, move_spline, target_spline, translate_spline)

class Pitch:
    values = None
    time_points = None
    contour = None
    parent = None
    glottal_closure_instants = None
    jitter = None
    simmer = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):

        return copy.deepcopy(self)

    def interpolate(self):
        f0 = self.values
        tf0 = self.time_points

        f0Plain = np.array([x for x in f0 if x > 0])
        t0Plain = np.array([tf0[x] for x in range(len(f0)) if f0[x] > 0])
        
        if f0Plain != []:
            f1 = scintp.interp1d(t0Plain, f0Plain, 'linear', fill_value=0,
                                 bounds_error=False)
            f0int = f1(tf0)
    
            idx_up = [x for x in range(len(tf0)) if tf0[x] <= t0Plain[0]]
            idx_down = [x for x in range(len(tf0)) if tf0[x] >= t0Plain[-1]]
            f0int[idx_up] = f0Plain[0]
            f0int[idx_down] = f0Plain[-1]            
        else:
            f0int = f0
            
        self.values = f0int
        return f0int
    
    def timeinterpolate(self, time_out, minf0=70):
        y = self.values
        tf0 = self.time_points
        
        f1 = scintp.interp1d(tf0,y,kind='linear',bounds_error=False, fill_value=0)
        y_out = f1(time_out)
        y_out[y_out<=minf0] = 0

        self.values = y_out
        self.time_points = time_out

    def getcontour(self, len_win=0.3):

        f0 = self.values
        tf0 = self.time_points

        srF0 = 1/(tf0[2]-tf0[1])
        nwin = int(len_win * srF0)
        if np.mod(nwin, 2) > 0:
            nwin = nwin + 1
        novlp = int(nwin * 1 / 2)
        nhop = nwin-novlp
        zeros2add = np.zeros((nhop, 1))

        f0up = np.vstack((zeros2add, f0.reshape(-1, 1), zeros2add))
        f0up = f0up.squeeze()

        w = np.ones(nwin)
        f0Out = np.zeros_like(f0up)

        k = 0
        isC = True
        while isC:
            deb = int(k * nhop)
            fin = deb + nwin - 1
            if fin > len(f0up) - 1:
                fin = len(f0up) -1
                isC = False
            idx = np.arange(deb, fin+1).astype(int)
            xTmp = f0up[idx]

            pp = np.polyfit(idx, xTmp, 2)
            newPitch = np.polyval(pp, idx)
            f0Out[idx] += f0Out[idx] + (newPitch[:len(idx)] * w[:len(idx)])
            k += 1

        f0Out = f0Out / 3
        f0Out = f0Out[nhop:-nhop]

        winOrd = int(0.3 * srF0)
        f0Out = sm.nonparametric.lowess(f0Out,
                                        np.arange(len(f0Out)),
                                        frac=winOrd/len(f0Out),
                                        return_sorted=False)

        self.contour = Contour('values', f0Out,
                               'time_points', self.time_points,
                               'parent', self)

    def plot(self):

        if self.contour is None:
            self.getcontour()
            
        if self.contour.spline is not None:
            spl_tp = self.contour.spline.time_points
            spl_v = self.contour.spline.values
        else:
            spl_tp = []
            spl_v = []
            
        return pitch_plot(self.time_points, self.values, self.contour.values,
                          spl_tp, spl_v, self.parent)

    def shift(self, factor):
        self.values = self.values * factor
        if self.contour is not None:
            self.contour.values = self.contour.values * factor
        if self.contour.normalized is not None:
            self.contour.detrend()
        if self.contour.spline is not None:
            self.contour.getSpline()
            self.contour.spline.parent = self

    def changespan(self, factor):
        f0 = self.values

        if self.contour is None:
            self.getcontour()

        contIn = self.contour.values
        if self.contour.normalized is None:
            self.contour.detrend()
        self.contour.normalized.values = self.contour.normalized.values * factor
        self.contour.retrend()

        self.values = (f0 - contIn) + self.contour.values
        self.getcontour()
        if self.contour.normalized is not None:
            self.contour.detrend()
        if self.contour.spline is not None:
            self.contour.getspline()
            self.contour.spline.parent = self

    def flattening(self):
        f0 = self.values

        if self.contour is None:
            self.getcontour()
        meanPitch = self.contour.values

        self.values = f0 - meanPitch + np.mean(f0)
        self.getcontour()
        if self.contour.normalized is not None:
            self.contour.detrend()
        if self.contour.spline is not None:
            self.contour.getSpline()
            self.contour.spline.parent = self

    def changecontour(self, target):

        if self.contour is None:
            self.getcontour()
        self.values = (self.values - self.contour.values +
                       target.contour.values)
        self.contour = target.contour.copy()
        self.contour.parent = self
        if self.contour.spline is not None:
            self.contour.getSpline()
            self.contour.spline.parent = self

    def changedeclination(self, target):

        if self.contour is None:
            self.getContour()
        if self.contour.normalized is None:
            self.contour.detrend()

        x = np.arange(len(self.values))
        pDecl = [np.tan(target), 0]
        pitchTrend = np.polyval(pDecl, x)
        newCont = self.contour.normalized.values + pitchTrend + self.contour.mean
        self.values = self.values - self.contour.values + newCont
        self.contour.values = newCont
        if self.contour.spline is not None:
            self.contour.getSpline()
            self.contour.spline.parent = self

class Contour(Pitch):
    mean = None
    normalized = None
    max_peaks = None
    min_peaks = None
    locs_max = None
    locs_min = None
    declination = None
    span = None
    peak_extent = None
    local_peak_dynamic = None
    top_angle = None
    base_angle = None
    spline = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def getspline(self, len_win=0.15):

        tf0 = self.time_points
        f0 = self.values
        
        srOut = 1/len_win

        srIn = 1/(tf0[2]-tf0[1])
        D = int(srIn / srOut)
        if D > 1:
            newTime = tf0[::D]
            newPitch = f0[::D]
        else:
            newTime = tf0
            newPitch = f0
        spl = scintp.splrep(newTime, newPitch)
        splFreq = scintp.splev(tf0, spl)
        (t, c, k) = spl

        splineObj = Spline('time_points', tf0,
                           'values', splFreq,
                           'knots', t,
                           'coeff', c,
                           'order', k,
                           'parent', self)

        self.spline = splineObj

        return splineObj
    
    def check_spline(self):
        if self.spline is None:
            self.getspline()

    def changespline(self, pts, factor, target=False):

        self.check_spline()
        pts = check_list(pts)
        if target:  
            target_spline(self.spline, pts, factor) 
        else:
            translate_spline(self.spline, pts, factor)

        if hasattr(self.parent, "values"):
            self.parent.values = (self.parent.values - self.parent.contour.values +
                              self.spline.values)
        self.values = self.spline.values

    def shiftspline(self, pts, factor):

        self.check_spline()
        pts = check_list(pts)
        move_spline(self.spline, pts, factor)

        self.values = self.spline.values

    def getlocalmax(self):

        f0 = self.values
        df0 = np.diff(f0)
        proddf0 = np.sign(df0[:-1]*df0[1:])
        locs = [x+1 for x in range(len(proddf0)) if proddf0[x] < 0]
        maxlocs = [x for x in locs if df0[x] < 0]
        minlocs = [x for x in locs if df0[x] > 0]

        self.max_peaks = f0[maxlocs]
        self.min_peaks = f0[minlocs]
        self.locs_max = maxlocs
        self.locs_min = minlocs

    def getbaselines(self):

        if (self.max_peaks is None or
                self.min_peaks is None or
                self.locs_max is None or
                self.locs_min is None):
                self.getlocalmax()

        maxpks = self.max_peaks
        minpks = self.min_peaks
        maxlocs = self.locs_max
        minlocs = self.locs_min


        warnings.simplefilter('ignore', np.RankWarning)
        pTop = np.polyfit(maxlocs, maxpks, 1)
        angleTop = np.arctan(pTop[0])
        
        pBase = np.polyfit(minlocs, minpks, 1)
        angleBase = np.arctan(pBase[0])
        self.declination = (angleTop - angleBase) / 2
        self.top_angle = angleTop
        self.base_angle = angleBase

    def detrend(self):

        if self.declination is None:
            self.getbaselines()

        x = np.arange(len(self.values))
        pDecl = [np.tan(self.declination), 0]
        pitchTrend = np.polyval(pDecl, x)
        detrend = self.values - pitchTrend
        self.mean = np.mean(detrend)
        self.normalized = Contour('values', detrend - np.mean(detrend),
                                  'time_points', self.time_points)

    def retrend(self):

        if self.declination is None:
            self.getbaselines()

        x = np.arange(len(self.values))
        pDecl = [np.tan(self.declination), 0]
        pitchTrend = np.polyval(pDecl, x)
        self.values = self.normalized.values + pitchTrend + self.mean

    def getpitchspan(self):
        if self.normalized is None:
            self.detrend()
        if self.normalized.max_peaks is None or self.normalized.min_peaks is None:
            self.normalized.getlocalmax()

        maxpks = self.normalized.max_peaks
        minpks = self.normalized.min_peaks
        self.span = np.mean(maxpks) - np.mean(minpks)

        return self.span

    def getnppe(self):
        if self.normalized is None:
            self.detrend()
        if self.normalized.max_peaks is None:
            self.normalized.getlocalmax()

        maxpks = self.normalized.max_peaks
        self.peak_extent = np.max(maxpks)

        return self.peak_extent

    def getlpdyn(self):
        if self.normalized is None:
            self.detrend()
        if (self.normalized.max_peaks is None or
                self.normalized.min_peaks is None or
                self.normalized.locs_max is None or
                self.normalized.locs_min is None):

            self.normalized.getlocalmax()

        tf0 = self.time_points
        maxpks = np.array(self.max_peaks).reshape(-1, 1)
        minpks = np.array(self.min_peaks).reshape(-1, 1)
        maxlocs = np.array(self.locs_max).reshape(-1, 1)
        minlocs = np.array(self.locs_min).reshape(-1, 1)

        localMaxima = np.vstack((maxpks, minpks)).squeeze()
        locsMaxima = np.vstack((maxlocs, minlocs)).squeeze()
        idxSort = np.argsort(locsMaxima)
        localMaxima = localMaxima[idxSort]
        locsMaxima = locsMaxima[idxSort]

        df = np.abs(localMaxima[1:] - localMaxima[:-1])
        dt = tf0[locsMaxima[1:]] - tf0[locsMaxima[:-1]]

        self.local_peak_dynamic = np.mean(df / dt)

        return self.local_peak_dynamic

    def db2lin(self):
        return  (10**(self.values / 10)) * 2e-5
    def lin2db(self):
        return  10 * np.log10(self.values / 2e-5)

class Spline(Contour):
    parent = None
    coeff = None
    knots = None
    order = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def interpolate(self):

        spl = (self.knots, self.coeff, self.order)
        splFreq = scintp.splev(self.parent.time_points, spl)

        self.values = splFreq

class Intensity(Contour):

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def interpolate(self, SpObj=None):

        if SpObj is None:
            SpObj = self.parent
        y = SpObj.signal
        sr = SpObj.sampling_frequency

        txx = np.arange(len(y)) / sr
        f1 = scintp.interp1d(self.time_points, self.values,
                             'linear', fill_value="extrapolate")
        contIntp = f1(txx)

        intOut = self.copy()
        intOut.values = contIntp
        intOut.time_points = txx
        if self.contour is not None:
           f1 = scintp.interp1d(self.time_points, self.contour.values,
                                'linear', fill_value="extrapolate")
           intOut.contour.values = f1(txx)
        if self.normalized is not None:
            intOut.getnormalized()
        if self.spline is not None:
            intOut.getspline()

        return intOut

    def getcontour(self, len_win=0.15):

        x = self.values
        tVec = self.time_points

        sr = 1/(tVec[2]-tVec[1])

        winOrd = int(len_win * sr)
        xOut = sm.nonparametric.lowess(x, np.arange(len(x)),
                                       frac=winOrd/len(x),
                                       return_sorted=False)
        self.contour = Contour('values', xOut,
                               'time_points', self.time_points,
                               'parent', self)

    def getnormalized(self):

        if self.contour is None:
            self.getcontour()

        ObjOut = self.contour.copy()
        ObjOut.values = self.values - self.contour.values
        self.normalized = ObjOut

    def plot(self):

        yS = self.parent.signal
        tVecSignal = np.arange(len(yS)) / self.parent.sampling_frequency
        yI = self.db2lin()

        return contour_plot(tVecSignal, yS, 
                            self.time_points, yI * np.max(yS) / np.max(yI),
                            parent=self.parent)

    def changespan(self, factor):

        if self.contour is None:
            self.getcontour()
        if self.normalized is None:
            self.getnormalized()

        self.values = self.normalized.values * factor + self.contour.values

    def flattening(self):
        x = self.values

        if self.normalized is None:
            self.detrend()

        self.mean = np.mean(x)
        self.values += self.mean



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jun  2 18:38:50 2019

@author: benjamin
"""

import json
import os
from os import path
import numpy as np
import scipy.signal as ssg
from ._set import miscstruct
from .oscillator import Glottis
from .vtnetwork import VtNetwork
from .vtwaveguide import VtWaveguide, MainOralTract, GlottalChink
from .areafunction import AreaFunction
from .utils import Cosine_interp, lininterp1

def importjson(jsonFile, seq=None):
    
    if seq is None:
        if path.isfile(jsonFile):
            with open(jsonFile, 'r') as myfile:
                jstr = myfile.read().replace('\n', '')
            idjson = json.loads(jstr)
            af = [x['area'] * 1e-4 for x in idjson['tubes']]
            lf = [x['x'] * 1e-2 for x in idjson['tubes']]
            vArea = [x['velumArea'] * 1e-4 for x in idjson['tubes']]
            vpo = (idjson['VelumOpeningInCm'] * 1e-2)**2
        else:
            print('Warning: file does not exist')
            return
    else:
        nFrame = len(seq)
        for kF in range(nFrame):
            fTmp = jsonFile + '_' + '%.4d' %(seq[kF]) + '.json'
            if path.isfile(fTmp):
                with open(fTmp, 'r') as myfile:
                    jstr = myfile.read().replace('\n', '')
                idjson = json.loads(jstr)
                afTmp = [x['area'] * 1e-4 for x in idjson['tubes']]
                lfTmp = [x['x'] * 1e-2 for x in idjson['tubes']]
                vAreaTmp = [x['velumArea'] * 1e-4 for x in idjson['tubes']]
                vpoTmp = (idjson['VelumOpeningInCm'] * 1e-2)**2
                if kF == 0:
                    nTubes = len(afTmp)
                    af = np.zeros((nTubes, nFrame))
                    lf = np.zeros((nTubes, nFrame))
                    vpo = np.zeros(nFrame)
                    vAreaTmp = np.zeros((nTubes, nFrame))
                af[:,kF] = afTmp
                lf[:,kF] = lfTmp
                vpo[kF] = vpoTmp
                vArea[:,kF] = vAreaTmp
                
    areaObj = AreaFunction('area', np.array(af),
                           'length', np.array(lf))

    if (np.sum(vpo) == 0) and (np.sum(vArea) == 0):
        obj_out = VtWaveguide('area_function', areaObj)
    else:
       obj_out = MainOralTract('area_function', areaObj,
            'velopharyngeal_port', np.array(vpo),
            'velum_area', np.array(vArea))
    areaObj.parent = obj_out
    return np.array(af), np.array(lf), obj_out

def vtjsonread(jsonfile, pathaf, prm=None):
    
    if prm is None:
        prm = miscstruct()

    if not hasattr(prm,'fsim'):
        prm.fsim = 6e4
    if not hasattr(prm, 'amin'):
        prm.amin = 1e-11
    if not hasattr(prm,'agmax'):
        prm.agmax = 4e-5
    if not hasattr(prm,'psubmax'):
        prm.psubmax = 700
    fsim = prm.fsim
    amin = prm.amin    
       
    VT = VtNetwork()
    with open(jsonfile, 'r') as myfile:
        jstr = myfile.read().replace('\n', '')
    idjson = json.loads(jstr)
    gs = idjson['globalScenario']
    nTubes = gs['numberOfTubes']
    
    ##### Area function
    af = gs['keyAreaFunctions']
    afId = [x.get('mainAreaFunctionFile') for x in af]
    naf = len(afId)
    afTp = [x.get('synthesisTime')*1e-3 for x in af]
    afTrans = [x.get('transitionFromPreviousAreaFunction') for x in af]
    tEnd = max(afTp)
    tVec = np.arange(0,tEnd+1/fsim,1/fsim)
    nPt = len(tVec)
    
    af0 = np.zeros((nTubes, naf))
    lf0 = np.zeros((nTubes, naf))
    varea0 = np.zeros((nTubes, naf))
    vpo0 = np.zeros(naf)
    
    afOut = np.zeros((nTubes, nPt))
    lfOut = np.zeros((nTubes, nPt))
    vpoOut = np.zeros(nPt)
    vareaOut = np.zeros((nTubes, nPt))
    
    for kaf in range(naf):   
        tmpFile = afId[kaf]
        idx = tmpFile.find('.')
        tmpList = os.listdir(pathaf)
        
        ff = [ x for x in tmpList if x.find(tmpFile[0:idx]) >= 0]
        ftmp = os.path.join(pathaf, ff[0])        
        with open(ftmp, 'r') as myfile:
            jstr2 = myfile.read().replace('\n', '')
            idtmp = json.loads(jstr2)
        vpo0[kaf] = (idtmp['VelumOpeningInCm']/1e2)**2 
        af0[:,kaf] = [x.get('area')*1e-4 for x in idtmp['tubes']]
        lf0[:,kaf] = [x.get('x')*1e-2 for x in idtmp['tubes']]
        varea0[:,kaf] = [x.get('velumArea')*1e-4 for x in idtmp['tubes']]
        
    
        if (kaf > 0):
            ind = np.argwhere((tVec>= afTp[kaf-1]) * (tVec <= afTp[kaf]))[:,0]
            xi = np.array([afTp[kaf-1], afTp[kaf]])
            xo = tVec[ind]
            yiAf = np.concatenate((af0[:,kaf-1].reshape(-1,1), af0[:,kaf].reshape(-1,1)),axis=1)
            yiLf = np.concatenate((lf0[:,kaf-1].reshape(-1,1), lf0[:,kaf].reshape(-1,1)),axis=1)
            yiVaO = np.concatenate((varea0[:,kaf-1].reshape(-1,1), varea0[:,kaf].reshape(-1,1)),axis=1)
            yiVpO = np.array([vpo0[kaf-1], vpo0[kaf]]).reshape(-1,1).T
    
            if afTrans[kaf] == 'COS':        
                afOut[:,ind] = Cosine_interp( xi, yiAf, xo )
                lfOut[:,ind] = Cosine_interp( xi, yiLf, xo )
                vareaOut[:,ind] = Cosine_interp( xi, yiVaO, xo )
                vpoOut[ind] = Cosine_interp( xi, yiVpO, xo )
            elif afTrans[kaf] == 'LIN':
                afOut[:,ind] = lininterp1( xi, yiAf, xo, 1 )
                lfOut[:,ind] = lininterp1( xi, yiLf, xo, 1)
                vareaOut[:,ind] = lininterp1( xi, yiVaO, xo, 1)
                vpoOut[ind] = lininterp1( xi, yiVpO, xo )
      
    
    afOut = afOut-vareaOut
    ##### Check null values
    afOut[afOut < amin] = amin
    lfOut[lfOut < amin] = amin
    vareaOut[vareaOut < amin] = amin
    vpoOut[vpoOut < amin] = amin
    
    
    MOT = MainOralTract()
    MOT.area_function = AreaFunction('area', afOut,
                                     'length', lfOut, 
                                     'parent', MOT)
    if VT.waveguide is None:
        VT.waveguide = []
    VT.waveguide.append(MOT)
    ###### Check for nasal coupling and build nasal tract if necessary
    if np.sum(vpo0) > 0:
        kn = np.zeros(nPt)
        for k in range(nPt):
            tmp = list(vareaOut[:,k]) 
            idx = [x for x in range(len(tmp)) if tmp[x] > 0][0] #list(vareaOut[:,k]).index(not 0)
            if idx != []:
                kn[k] = idx
    
        kn = np.ceil(np.median(kn))
    
    #### F0
    F0 = gs['F0Commands']
    
    F0Hz = [x.get('FOInHz') for x in F0]
    F0tp = [x.get('timeInMs')*1e-3 for x in F0]
    idx0 = [x for x in range(len(F0Hz)) if F0Hz[x]>0]
    F0Hz = [F0Hz[x] for x in idx0]
    F0tp = [F0tp[x] for x in idx0]
    
    F0Out = lininterp1(F0tp,F0Hz,tVec)
    
    ##### Ag
    ag = gs['keyAG0']
    ag0 = [x.get('Ag0Opening') for x in ag]
    agPhon = [x.get('phoneticLabel') for x in ag]
    agTp = [x.get('synthesisTime')*1e-3 for x in ag]
    agTrans = [x.get('transitionFromPreviousAreaFunction') for x in ag]
    nag = len(ag0)
    
    agOut = np.zeros(nPt)    
    for kag in range(nag):
        if (kag > 0):
            ind = np.argwhere((tVec>= agTp[kag-1]) * (tVec <= agTp[kag]))[:,0]       
            if agTrans[kag] == 'COS':
                agOut[ind] = Cosine_interp( np.array([ agTp[kag-1], agTp[kag] ]), np.array([ag0[kag-1], ag0[kag] ]).reshape(-1,1).T, tVec[ind] )
            elif agTrans[kag] == 'LIN':
                agOut[ind] = lininterp1( np.array([ agTp[kag-1], agTp[kag] ]), np.array([ ag0[kag-1], af0[kag] ]).reshape(-1,1).T, tVec[ind] )
    
    Gl = Glottis()
    Gl.fundamental_frequency = F0Out
    lch = np.sqrt(agOut)
    lch[lch <= amin] = amin # if length = 0 => set to the minimal value
    lch[lch >= 1] = 1-amin #  if length = 1 => set to the maximal value (1-vmin)
    Gl.partial_abduction = lch
    
    ##### Check existence of glottal opening and build glottal chink if necessary
    if np.sum(agOut) > 0:
        agOut = agOut*prm.agmax
        agOut[agOut<amin] = amin
        Chink = GlottalChink()
        Chink.parent_wvg = [MOT]
        achTmp = agOut.reshape(1,-1)
        lchTmp = Gl.x_position[-1] * np.ones_like(achTmp)
        Chink.area_function = AreaFunction('area', achTmp,
                                           'length', lchTmp, 
                                           'parent', Chink)
        
        VT.waveguide.append(Chink)
    
    if VT.oscillators is None:
        VT.oscillators = []
    VT.oscillators.append(Gl)
    
    #### Segmentation
    seg = gs['segmentation']
    segStart = np.array([x.get('beginsAtInMs')*1e-3 for x in seg])
    segEnd = np.array([x.get('endsAtInMs')*1e-3 for x in seg])
    segPhon = [x.get('sampaCode') for x in seg]
    
    #### Build subglottal pressure
    psubOut = np.ones(nPt) * prm.psubmax     
    phonDur = (segEnd - segStart)
    nup = int(0.025 * prm.fsim)
    ndown = int(0.1 * prm.fsim)
    
    for k in range(len(segPhon)):
        ind = [x for x in range(len(tVec)) if tVec[x] >= segStart[k] and tVec[x] <= segEnd[k]]
        if segPhon[k] == '#':
            psubOut[ind] = 0
            if k < len(segPhon) - 1:
                if segPhon[k+1] != '#':
                    ind2ch = np.arange(ind[-1] - nup, ind[-1] + nup)
                    w = ssg.hann(2 * len(ind2ch))
                    w = w[:len(ind2ch)]
                    psubOut[ind2ch] = w * prm.psubmax
        else:
            if k < len(segPhon) - 1:
                if segPhon[k+1] == '#':
                    ind2ch = np.arange(ind[-1] - ndown, ind[-1] + ndown)
                    ind2ch = [x for x in ind2ch if x < len(psubOut)]
                    w = ssg.hann(2 * len(ind2ch))
                    w = w[:len(ind2ch)][::-1]
                    psubOut[ind2ch] = w * prm.psubmax
 
    VT.subglottal_control = psubOut
    VT.phonetic_label = segPhon
    VT.phonetic_instant = np.concatenate((segStart.reshape(-1,1), segEnd.reshape(-1,1)),axis=1)
    VT.simulation_frequency = fsim

    return VT


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 14 09:33:59 2018

@author: benjamin
"""
import numpy as np
   
class struct:
    def __init__(self):
        Tair = 37
        self.rho = 1.2929 * (1 - 0.0037 * Tair)
        self.c_s = 331.45 * (1 + 0.0018 * Tair)
        self.mu = 1.85e-5
        self.a = 130 * np.pi
        self.b = (30 * np.pi)**2
        self.c1 = 4
        self.wo2 = (406 * np.pi)**2
        self.c1n = 72
        self.heat_cond = 0.0034
        self.specific_heat = 160
        self.adiabatic = 1.4
        self.wallyield = True
        self.loss = True
        self.freq = np.arange(0,5050,50,dtype='float') #Vector of frequencies
        self.nform = 4
        self.length_var = True
        self.lips_var = True
        self.area_weight = 0
        self.length_weight = 0;
        self.potential_weight = 0
        self.threshold = 1
        self.maxNumIter = 1e3
        self.psi = 1
 
class timestruct:
    def __init__(self):
        Tair = 37
        ######### Coefficients for wall parameters
        loss_terms = 'birk' # maeda, birk or mokh
        if loss_terms ==  'maeda': # ## From Maeda
            wall_resi = 1600 #/* wall resistance, gm/s/cm2		*/
            wall_mass = 1.5	#/* wall mass per area, gm/cm2		*/
            wall_comp = 3.0e5	#/* wall compliance			*/
        if loss_terms ==  'birk': # #### From Birkholz
            wall_resi = 8e3	#/* wall resistance, gm/s/cm2		*/
            wall_mass = 21e0	#/* wall mass per area, gm/cm2		*/
            wall_comp = 8.45e6	#/* wall compliance			*/
        if loss_terms ==  'mokh': #### From Mokhtari
            wall_resi = 14e3	#/* wall resistance, gm/s/cm2		*/
            wall_mass = 15e0	#/* wall mass per area, gm/cm2		*/
            wall_comp = 3e6	#/* wall compliance			*/
        
        self.fs = 2e4
        self.T = 1/self.fs
        self.tm = 1
        self.tcl = 0.01
        self.namp = 1e-7
        self.dispcomp = True
        self.wallyield = True
        self.dynterm = True
        self.wr = wall_resi
        self.wm = wall_mass
        self.wc = wall_comp
        self.rho = 1.2929 * (1 - 0.0037 * Tair)
        self.c = 331.45 * (1 + 0.0018 * Tair)
        self.c_s = 331.45 * (1 + 0.0018 * Tair)
        self.mu = 1.85e-5
        self.lg = 0.01
        self.Xg = 0.3e-2
        self.kc = 1.42
        self.dist_source = 2 # Gaussian noise
        self.fh = 0.2e3 # Cut-off frequency for high-pass filter
        self.fl = 4e3 # Cut-off frequency for low-pass filter
        self.ordh = 1 # Order of high-pass filter
        self.ordl = 7 # Order of low-pass filter
        self.bf = np.array([0.9, 0.1, -0.8] )
        self.af = 1
        self.K_b = self.kc*self.rho
        self.Lord = 50e-3
        self.Rec = 1.7e3 # Reynolds threshold for frication noise generation
        self.G = 9 / 128
        self.S = 3 / 8
        self.sep = 1.2
        self.filt = 'fir1'
        self.blim = 0
        self.beta = 5e-8
        self.nmass = 2
        self.amin = 1e-11
        self.model = 'ishi'
        self.rmlt = 4
        self.a = 130 * np.pi
        self.b = (30 * np.pi)**2
        self.c1 = 4
        self.wo2 = (406 * np.pi)**2
        self.c1n = 72
        self.heat_cond = 0.0034
        self.specific_heat = 160
        self.adiabatic = 1.4
        self.wallyield = True
        self.loss = True
        self.freq = np.arange(0,5050,50,dtype='float') #Vector of frequencies
        self.nform = 4
        self.length_var = True
        self.lips_var = True
        self.area_weight = 0
        self.length_weight = 0;
        self.potential_weight = 0
        self.threshold = 1
        self.maxNumIter = 1e3
        self.psi = 1
        
        
class miscstruct:
    def __init__(self):
        self.dum = 0
